static char rcsid[] = "@(#)$Id: remote_mbx.c,v 1.22 2001/03/09 17:55:01 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.22 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************/

#include "headers.h"
#include "mbx_imp.h"
#include "s_me.h"
#include "s_elm.h"

#ifdef REMOTE_MBX

/* Seems that h_errno is macro on AIX */
#ifndef h_errno
extern int h_errno;
#endif

#include <errno.h>
extern int errno;

void zero_remote_account(ra)
     struct remote_account *ra;
{
    /* bzero is defined hdrs/defs.h */
    bzero (ra, sizeof (struct remote_account));

    ra->hostaddr.sa.sa_family = AF_UNSPEC;
    ra->socket                = -1;
    ra->username              = NULL;
    ra->host                  = NULL;
}

/* Closed only implicity on exit */
FILE * transaction_file = NULL;

int set_transaction_file(filename)
     CONST char *filename;
{
    FILE * f;

    if (getuid() != geteuid() || getgid() != getegid()) {
	lib_error(FRM("%s: Dropping privilges\n"),filename);
	if (-1 == setgid(getgid()) || -1 == setuid(getuid())) {
	    int err = errno;
	    lib_error(FRM("%s: setgid/setuid failed: %s\n"),
		      filename,error_description(err));
	    return 0;
	}
    }

    f = fopen(filename,"a");
    if (!f) {
	int err = errno;
	lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotWriteable,
			  "File %.50s is not writeable: %s"),
		  filename, error_description(err));
	return 0;
    }

    if (transaction_file) {
	fprintf(transaction_file,
		"\n===== Changing logging to file %s\n",
		filename);
	fclose(transaction_file);
    }
    transaction_file = f;
    
#ifdef SETLINEBUF
    setlinebuf(transaction_file);
#endif

    chmod(filename,0600);
    return 1;
}

void free_remote_account(ra)
     struct remote_account *ra;
{
    ra->hostaddr.sa.sa_family = AF_UNSPEC;

    if (-1 != ra->socket) {
	clear_action(ra->socket); /* Assurance */
	close(ra->socket);
	ra->socket = -1;
    }
    if (ra->username) {
	free(ra->username);
	ra->username = NULL;
    }
    if (ra->host) {
	free(ra->host);
	ra->host = NULL;
    }
}


int connect_remote_account(ra,portlist,got,he)
     struct remote_account *ra;
     PORTS portlist[];
     PORTS *got;
     struct hostent *he;	    
{
    int ok = 0;
    int last_error = 0;
    
#ifdef I_NETINET_IN
    if (he->h_addrtype == AF_INET) {
	int idx;

	ra->hostaddr.sin.sin_family  = he->h_addrtype;

	if (sizeof(ra->hostaddr.sin.sin_addr)  != he->h_length) {
	    lib_error(FRM("%s: Bad IP-addr length %d (should be %d)"),
		      he->h_name,he->h_length,
		      sizeof(ra->hostaddr.sin.sin_addr));
	    ok = 0;
	    goto clean;
	}

	if(-1 != ra->socket) {
	    dprint(12, (debugfile,
			"connect_remote_account: Closing fd=%d for reopening\n",
			ra->socket));
	    close(ra->socket);
	    ra->socket = -1;
	}

	ra->socket = socket(he->h_addrtype,SOCK_STREAM,IPPROTO_TCP);
	if (-1 == ra->socket) {
	    int err =errno;
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeFailedToCreate,
			      "Failed to create socket: %s"),
		      error_description(err));
	    ok = 0;
	    goto clean;
	}

	for (idx = 0; !ok  && he->h_addr_list[idx]; idx++) {
	    int idx2;
	    memcpy(&(ra->hostaddr.sin.sin_addr),
		   he->h_addr_list[idx],he->h_length);

	    lib_transient(CATGETS(elm_msg_cat, MeSet,MeConnecting,
				  "Connecting to %s [%s]... (%d)"),
			  he->h_name,
			  inet_ntoa(ra->hostaddr.sin.sin_addr),
			  idx);


	    for (idx2 = 0; portlist[idx2] != PORT_end && !ok; idx2++) {
		ra->hostaddr.sin.sin_port = htons(portlist[idx2]); 
		
#ifdef NEED_REOPEN_AFTER_FAILED_CONNECT
		if (idx || idx2) {
		    dprint(12, (debugfile,
				"connect_remote_account: NEED_REOPEN_AFTER_FAILED_CONNECT -- reopening...\n"));
		    close(ra->socket);
		    ra->socket = socket(he->h_addrtype,
					SOCK_STREAM,IPPROTO_TCP);
		    if (-1 == ra->socket) {
			dprint(12, (debugfile,
				    "connect_remote_account: socket failed: %s\n",
				    error_description(errno)));
			goto fail_connect;
		    }			
		}
#endif

		if (-1 == connect(ra->socket,& (ra->hostaddr.sa),
				  sizeof (ra->hostaddr.sin))) {
		    last_error= errno;
		    dprint(12, (debugfile,
				"connect_remote_account: connect failed: %s\n",
				error_description(last_error)));
		    
		    if (errno == EINVAL) {
			/* Seems that it is not not allowed several connect 
			 * attempts with same socket ...
			 */
			last_error= errno;
			dprint(12, (debugfile,
				    "connect_remote_account: -- reopening socket and retrying...\n"));
			
			close(ra->socket);
			ra->socket = socket(he->h_addrtype,SOCK_STREAM,
					    IPPROTO_TCP);
			if (-1 == ra->socket ||
			    -1 == connect(ra->socket,&(ra->hostaddr.sa),
					  sizeof (ra->hostaddr.sin))) {
			    last_error= errno;
			    dprint(12, (debugfile,
					"connect_remote_account: connect failed: %s\n",
					error_description(last_error)));
			} else
			    ok = 1;
		    }
		} else
		    ok = 1;
		
		/* If got connection to host and and it refused
		   connection, try another ports
		*/
		if (!ok && last_error != ECONNREFUSED) 
		    break;			   
		if (ok) {
		    *got = portlist[idx2];
		    
		    dprint(12, (debugfile,
				"connect_remote_account: Connection succeed %s [%s], port %d\n",
				he->h_name,
				inet_ntoa(ra->hostaddr.sin.sin_addr),
				portlist[idx2]));

		    if (transaction_file) {
			time_t tmp = time(NULL);
			struct tm * zz = localtime(&tmp);
			fprintf(transaction_file,
				"%d [%d] %02d:%02d:%02d === CONNECT %s [%s], port %d\n",
				getpid(),ra->socket,
				zz ? zz->tm_hour : 00,
				zz ? zz->tm_min  : 00,
				zz ? zz->tm_sec  : 00,
				he->h_name,
				inet_ntoa(ra->hostaddr.sin.sin_addr),
				portlist[idx2]);
		    }
		}
	    }	    
	}
    } else 
#endif /* I_NETINET_IN */	
	/* Should AF_INET6 also be supported on here? ... */
	{
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeUnsupportedAddrType,
			      "Name %s have odd type address"),
		      he->h_name);
	    ok = 0;
	    goto clean;
	}

    if (!ok) {
    fail_connect:
	lib_error(CATGETS(elm_msg_cat, MeSet,MeConnectFailed,
			  "Failed to connect %s: %s"),
		  he->h_name,error_description(last_error));
	ok = 0;
	goto clean;
    }

    /* We set folder to non-blocking after connect, because
     * non-blocking connect is little complicate to use
     */
    
    if (-1 == fcntl(ra->socket,F_SETFL,O_NONBLOCK)) {
	int err = errno;
	dprint(12, (debugfile,
		    "connect_remote_account: fcntl [O_NONBLOCK] failed: %s\n",
		    error_description(err)));
    }


 clean:
    dprint(12, (debugfile,
		"connect_remote_account=%d\n",ok)); 
    return ok;
}

/* -1 == name not found or bad syntax
   0 == not a remote address
   1 == name found 
*/
int split_remote_name(name,X,he,rest)
     char *name;
     struct remote_account *X;
     struct hostent **he;	    
     char **rest;
{
    char *sep;
    int ret = 0;
    
    dprint(12, (debugfile,
		"split_remote_name: name=%s\n",name));
    
    *rest = NULL;
    zero_remote_account(X);
    
    *he = NULL;

    *rest = strpbrk(name,"/:");
    sep  = strchr(name,'@');

    if (sep && (!*rest || *rest > sep)) {
	if (sep == name || sep[1] == '\0' ||
	    *rest == sep+1) {
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeBadRemoteMailbox,
			      "Bad remote mailbox: %s"),
		      name);
	    ret = -1;
	    goto clean;
	}
	
	if (*rest) {
	    CONST char * START = sep+1;
	    X->host = safe_malloc(*rest - START+1);
	    memcpy(X->host,START,*rest - START);
	    X->host[*rest - START] = '\0';
	} else
	    X->host = safe_strdup(sep+1);	   

	X->username = safe_malloc(sep - name+1);
	memcpy(X->username,name,sep - name);
	X->username[sep - name] = '\0';

	lib_transient(CATGETS(elm_msg_cat, MeSet,MeLookingUp,
			      "Looking up %s ..."),
		      X->host);

	*he = gethostbyname(X->host);
	if (!*he) {
	    dprint(12, (debugfile,
			"split_remote_name: %s: h_errno=%d\n",
			X->host,h_errno));
	    switch(h_errno) {
	    case HOST_NOT_FOUND:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeNoRemoteHost,
				  "No remote mailbox server host: %s"),
			  X->host);		
		break;
	    case NO_ADDRESS:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeNoAddress,
				  "Name %s have not IP-address"),
			  X->host);
		break;
	    case TRY_AGAIN:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeTemporary,
				  "Address for %s is not yet found..."),
			  X->host);
		break;			  
	    default:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeNameError,
				  "Failed to get address for %s"),
			  X->host);
		break;
	    }
	    ret = -1;
	    goto clean;
	}

	/* Canonify host name ... */
	free(X->host);
	X->host = safe_strdup((*he)->h_name);

	dprint(12, (debugfile,
		    "split_remote_name: *he=%p; h_name=%s, h_addrtype=%d\n",
		    *he,
		    (*he)->h_name ? (*he)->h_name : "<NULL>",
		    (*he)->h_addrtype));
	dprint(12, (debugfile,
		    "split_remote_name: username=%s, host=%s\n",
		    X->username ? X->username : "<NULL>",
		    X->host ? X->host : "<NULL>"));

	ret = 1;
    } 

 clean:
    dprint(12, (debugfile,
	       "split_remote_name=%d: *rest=%s\n",
		ret,
		*rest ? *rest : "<NULL>"));

    return ret;
}		

/* May free X (if succeed) or copy */
int make_remote_mbox(fh,X,he,rest,rewrite)
     struct folder_info *fh; 
     struct remote_account *X;
     struct hostent *he; 
     char * rest;
     int rewrite;
{
    struct remote_account *status = NULL;

    PORTS ports[] = { PORT_imap4, PORT_pop3, PORT_end };
    PORTS ports_imaponly[] = { PORT_imap4, PORT_end };
    char * imap_only = NULL;  /* Flag and name of imap folder ... */
    PORTS got;
    static struct connection_cache *CX;

    dprint(10, (debugfile,
		"make_remote_mbox: fh=%X, rest=%s, rewrite=%d\n",
		fh,rest ? rest : "<NULL>", rewrite));

    if (rewrite) {
	/* canonify folder name .... */
	if (rest) {
	    /* Inbox is case insensitive */
	    if (0 == istrcmp(":inbox",rest))
		rest = ":INBOX";
	    
	    imap_only = safe_strdup(rest);
	    free(fh->cur_folder_sys);
	    fh->cur_folder_sys = elm_message(FRM("%s@%s%s"),
					     X->username,he->h_name,imap_only); 
	    /* No support of international folder name son here! */
	    free_string(&(fh->cur_folder_disp));
	    fh->cur_folder_disp = format_string(FRM("%s@%s%s"),
						X->username,he->h_name,imap_only); 
	    
	    rest = NULL; /* is no longer valid pointer if 
			    rest was pointing to fh->cur_folder_sys
			 */
	} else {
	    free(fh->cur_folder_sys);
	    fh->cur_folder_sys = elm_message(FRM("%s@%s"),
					     X->username,he->h_name);
	    free_string(&(fh->cur_folder_disp));
	    fh->cur_folder_disp = format_string(FRM("%s@%s"),
						X->username,he->h_name);    
	}
    }

    /* Only IMAP connections are cached */
    CX = locate_from_cache(X->username,X->host,&IMAP_connection);
    
    if (CX) {
	if (!imap_only)
	    imap_only = safe_strdup(":INBOX");
	fh -> folder_type = IMAP_MBX;
	dprint(8, (debugfile,
		   "make_remote_mbox: IMAP mailbox (from cache)\n"));
	
	fh->folder_type->init_it(fh);
	folder_from_connection(CX,fh);
	
	status = &(fh->p->a.imap_mbx.Ch->C);
	
	if (':' == imap_only[0]) 
	    fh->p->a.imap_mbx.folder = safe_strdup(imap_only+1);
	else
	    fh->p->a.imap_mbx.folder = safe_strdup(imap_only);
	
	free_remote_account(X);
    } else {
	if (!connect_remote_account(X,
				    imap_only ? ports_imaponly : ports,
				    &got,he))
	    goto fail;
	
	switch (got) {
	case PORT_pop3:
	    fh -> folder_type = POP_MBX;
	    
	    dprint(8, (debugfile,
		       "make_remote_mbox: POP mailbox (succeed)\n"));
	    
	    fh->folder_type->init_it(fh);
	    free_remote_account(&(fh->p->a.pop_mbx.C));
	    fh->p->a.pop_mbx.C = *X;	    
	    status = &(fh->p->a.pop_mbx.C);
	    
	    break;
	case PORT_imap4:
	    if (!imap_only)
		imap_only = safe_strdup(":INBOX");
	    fh -> folder_type = IMAP_MBX;
	    
	    dprint(8, (debugfile,
		       "make_remote_mbox: IMAP mailbox (succeed)\n"));
	    
	    fh->folder_type->init_it(fh);
	    
	    if (!join_connection(fh->p->a.imap_mbx.Ch,X,CON_greeting)) 
		fh->p->a.imap_mbx.Ch->a.imap_con.imap_state = IMAP_error;
	    else
		fh->p->a.imap_mbx.Ch->a.imap_con.imap_state = IMAP_idle;
	    status = &(fh->p->a.imap_mbx.Ch->C);
	    
	    if (':' == imap_only[0]) 
		fh->p->a.imap_mbx.folder = safe_strdup(imap_only+1);
	    else
		fh->p->a.imap_mbx.folder = safe_strdup(imap_only);
	    break;
	}	
    }

    if (status) {
	/* In here we can use temp_dir instead of default_temp because
	 * that file does not act as session lock ...
	 */
	if (imap_only) {
	    char *Y = strrchr(imap_only,'/');
	    if (!Y)
		Y = imap_only;
	    elm_sfprintf(fh-> cur_tempfolder,
			 sizeof fh -> cur_tempfolder,
			 FRM("%s%s%s@%s:%s"),
			 temp_dir,temp_mbox,
			 status->username,
			 status->host,Y+1);	    
	} else
	    elm_sfprintf(fh-> cur_tempfolder,
			 sizeof fh -> cur_tempfolder,
			 FRM("%s%s%s@%s"),
			 temp_dir,temp_mbox,
			 status->username,status->host);
    } 
    
 fail:

    if (imap_only)
	free(imap_only);
      
    if (status) {
	dprint(12, (debugfile,"make_remote_mbox=1\n"));
	return 1;
    }

    dprint(12, (debugfile,"make_remote_mbox=0\n"));
    return 0;
}

#endif /* REMOTE_MBX */


int remote_folder_type(fh)
     struct folder_info *fh;
{
    char * rest, *p;
    int ret = 0;
#ifdef REMOTE_MBX
    struct remote_account X;
    struct hostent *he = NULL;
    int z;
#endif

    /* Explicite disallow $MAIL to be interpreted as remote mailbox */

    if ('/' == fh->cur_folder_sys[0] ||
	((p = getenv("MAIL")) && strcmp(fh->cur_folder_sys, p) == 0) ||
	0 == access(fh->cur_folder_sys,ACCESS_EXISTS)) {
	dprint(8, (debugfile,
		   "remote_folder_type=0 (file: %s)\n",
		   fh->cur_folder_sys));
	return 0;
    }

#ifdef REMOTE_MBX
    
    /* -1 == name not found or bad syntax
        0 == not a remote address
	1 == name found 
    */
    if ((z = split_remote_name(fh->cur_folder_sys,&X,&he,&rest))) {
	
	if (z < 0) 
	    goto clean;

	/* May free X (if succeed) or copy */
	if (!make_remote_mbox(fh,&X,he,rest,1)) {
	clean:
	    free_remote_account(&X);
	    
	    fh -> folder_type = NO_NAME;
	    fh->folder_type->init_it(fh);
	}	           
	ret = 1;      /* Is 'remote mailbox' ... */	    
    }
#else
    dprint(9, (debugfile,
	       "remote_folder_type: (remote mailboxes are not supported.)\n"));
#endif /* REMOTE_MBX */

    dprint(8, (debugfile,
	       "remote_folder_type=%d\n",ret));
    return ret;
}

void zero_Read_Buffer(buffer)
     struct Read_Buffer *buffer;
{
    buffer->read_buffer = NULL;
    buffer->read_len    = 0;
}

void free_Read_Buffer(buffer)
     struct Read_Buffer *buffer;
{
    if (buffer->read_buffer) {
	free(buffer->read_buffer);
	buffer->read_buffer = NULL;
    }
    buffer->read_len = 0;
}

/* For non-lieral we read with quite short block because
   many answers are quite small
*/

#define READ_BLOCK   1024

int ReadFromSocket(fd,buffer,wanted)
     int fd; 
     struct Read_Buffer *buffer;
     int wanted;
{
    int n;

    if (wanted > 0) {
	buffer -> read_buffer = safe_realloc(buffer -> read_buffer,
					     buffer -> read_len + wanted);
	n = read(fd, buffer -> read_buffer + buffer -> read_len, wanted);
    } else {
	buffer -> read_buffer = safe_realloc(buffer -> read_buffer,
					     buffer -> read_len + READ_BLOCK);
	n = read(fd, buffer -> read_buffer + buffer -> read_len, READ_BLOCK);
    }

    if (n > 0) {

	if (transaction_file) {
	    time_t tmp = time(NULL);
	    struct tm * zz = localtime(&tmp);
	    int i;
	    int q = 0;

	    fprintf(transaction_file,
		    "%d [%d] %02d:%02d:%02d <<< [len %4d] ",
		    getpid(),fd,
		    zz ? zz->tm_hour : 00,
		    zz ? zz->tm_min  : 00,
		    zz ? zz->tm_sec  : 00,
		    n);

	    for (i = 0; i < n; i++) {
		unsigned char c = 
		    (unsigned char) buffer -> read_buffer[buffer -> read_len 
							 + i];
		
		if (isascii(c) && isprint(c) && c != '"') {
		    if (!q) {
			putc(' ',transaction_file);
			putc('"',transaction_file);
			q++;
		    }
		    putc(c,transaction_file);
		} else {
		    if (q) {
			putc('"',transaction_file);
			q = 0;
		    }
		    fprintf(transaction_file," %02X",c);

		    if (c == 0x0A && i < n-1) {
			fprintf(transaction_file,
				"\n%d [%d] %02d:%02d:%02d <<< [continue] ",
				getpid(),fd,
				zz ? zz->tm_hour : 00,
				zz ? zz->tm_min  : 00,
				zz ? zz->tm_sec  : 00);
		    }
		}		
	    }
	    if (q) {
		putc('"',transaction_file);
	    }
	    putc('\n',transaction_file);
	}

	buffer -> read_len += n;
	dprint(13, (debugfile,
                    "ReadFromSocket: Readed %d bytes (%d total)\n",
		    n,buffer->read_len));
    }

    return n;
}

int find_crlf(buffer, add_null)
     struct Read_Buffer *buffer;
     int add_null;
{
    char * p = buffer->read_buffer;
    int i;

    for (i = 0; i < buffer->read_len -1; i++)
	if ('\r' == p[i] && '\n' == p[i+1]) {
	    if (add_null) {
		p[i] = '\0';
		p[i+1] = '\0';
	    }
	    dprint(49, (debugfile,"find_crlf=%d (%d total)\n",
			i+2,buffer->read_len));
	    return i+2;
	}
    dprint(49, (debugfile,"find_crlf=0 (%d total)\n",
		buffer->read_len));
    return 0;
}

void cut_line(buffer, len)
     struct Read_Buffer *buffer;
     int len;
{
    char * p = buffer->read_buffer;
    
    buffer->read_len -= len;
    dprint(49, (debugfile,"cut_line: %d chars consumed, %d left\n",
		len,buffer->read_len));
    if (buffer->read_len)
	memmove(p, p+len, buffer->read_len);
}     


#ifdef REMOTE_MBX

static struct connection_cache * CACHE_LIST = NULL;

struct connection_cache * locate_from_cache(username,host,con_type)
     CONST char * username;
     CONST char * host;
     CONST struct connection_type  *con_type;
{
    struct connection_cache * ptr;
 
    for (ptr = CACHE_LIST; ptr; ptr = ptr -> next) {
	if (!ptr->C.username || !ptr->C.host) {
	    dprint(11, (debugfile,
			"locate_from_cache: %p: No host or username on cache!\n",
			ptr));
	    continue;
	}
	
	if (0 == strcmp(username,ptr->C.username) &&
	    0 == istrcmp(host,ptr->C.host) &&
	    (!con_type || con_type == ptr->type))
	    return ptr;
    }
   
    return NULL;
}

static void remove_from_cache P_((struct connection_cache *con,
				  int maybe));
static void remove_from_cache(con,maybe)
     struct connection_cache *con;
     int maybe;
{
    struct connection_cache * ptr, *prev = NULL;

    for (ptr = CACHE_LIST; ptr; prev=ptr, ptr = ptr -> next) {
	if (ptr == con) {
	    if (NULL == prev)
		CACHE_LIST = ptr->next;
	    else
		prev->next = ptr->next;
	    ptr->next = NULL;

	    dprint(11, (debugfile,
			"remove_from_cache: con=%p (%s@%s), type=%p: removed from cache\n",
			con,
			con->C.username ? con->C.username : "<NULL>",
			con->C.host ? con->C.host : "<NULL>",
			con->type));

	    return;
	}
    }
    con->next = NULL;
    
    if (maybe) {
	dprint(11, (debugfile,
		    "remove_from_cache: con=%p (%s@%s), type=%p: Not in cache\n",
		    con,
		    con->C.username ? con->C.username : "<NULL>",
		    con->C.host ? con->C.host : "<NULL>",
		    con->type));
	
    } else {
	dprint(1, (debugfile,
		   "remove_from_cache: con=%p (%s@%s), type=%p: Not in cache\n",
		   con,
		   con->C.username ? con->C.username : "<NULL>",
		   con->C.host ? con->C.host : "<NULL>",
		   con->type));
	
	panic("CONNECTION PANIC",__FILE__,__LINE__,"remove_from_cache",
	      "Connection not in connection cache",0);
    }
}

struct connection_cache *  create_connection(T)
     struct connection_type *T;
{
    struct connection_cache *ret;

    dprint(10, (debugfile,
		"create_connection: type=%p\n",
		T));

    ret = safe_malloc(sizeof (struct connection_cache));
    bzero((void *)ret,sizeof (struct connection_cache));

    ret->type  = T;
    ret->next  = NULL;
    ret->f     = NULL;
    ret->state = CON_error;

    zero_remote_account(&(ret->C));

    ret->type->cache_zero_it(ret);

    dprint(10, (debugfile,"create_connection=%p\n",ret));

    return ret;
}

void  free_connection(c)
     struct connection_cache **c;
{
    dprint(10, (debugfile,
		"free_connection: con=%p (%s@%s), type=%p\n",
		(*c),
		(*c)->C.username ? (*c)->C.username : "<NULL>",
		(*c)->C.host ? (*c)->C.host : "<NULL>",
		(*c)->type));

    if (CACHE_LIST)
	remove_from_cache((*c),1);

    if (-1 != (*c)->C.socket &&
	(*c)->state != CON_error)
	(*c)->type->cache_close_it((*c));

    (*c)->type->cache_free_it((*c));

    free_remote_account(&((*c)->C));

    (*c)->state = CON_error;
    (*c)->f    = NULL;
    (*c)->next = NULL;
    (*c)->type = NULL;

    free(*c);
    *c = NULL;
}

int  join_connection(c,X,st)
     struct connection_cache *c;
     struct remote_account *X;
     enum connection_state st;
{
    int ret = 0;

    dprint(10, (debugfile,
		"join_connection: con=%p, type=%p, X=%p (%s@%s), st=%d\n",
		c,
		c->type,
		X,
		X->username ? X->username : "<NULL>",
		X->host ? X->host : "<NULL>",
		st));

    free_remote_account(&(c->C));
    
    c->C     = *X;
    c->state = st;

    /* Avoid double free() ... */
    X->host      = NULL;
    X->username  = NULL;
    X->socket   = -1;

    ret = c->type->cache_open_it(c);

    dprint(10, (debugfile,"join_connection=%d  (state=%d)\n",
		ret,c->state));
    return ret;
}

int  login_connection(c)
     struct connection_cache *c;
{
    int ret = 0;

    dprint(10, (debugfile,
		"login_connection: con=%p (%s@%s), type=%p\n",
		c,
		c->C.username ? c->C.username : "<NULL>",
		c->C.host ? c->C.host : "<NULL>",
		c->type));

    if (c->state != CON_open) {
	dprint(10, (debugfile,
		    "login_connection: Opening it first...\n")); 
	ret = c->type->cache_open_it(c);
	if (!ret)
	    goto fail;
    }

    ret = c->type->cache_login_it(c);

 fail:
    dprint(10, (debugfile,"login_connection=%d\n",ret));
    return ret;
}

void folder_from_connection(c,f)
     struct connection_cache *c;
     struct folder_info *f;
{
    dprint(10, (debugfile,
		"folder_from_connection: con=%p (%s@%s), type=%p\n",
		c,
		c->C.username ? c->C.username : "<NULL>",
		c->C.host ? c->C.host : "<NULL>",
		c->type));
    dprint(10, (debugfile,
		"                      : folder=%p (%s), type=%p\n",
		f,f->cur_folder_sys,f -> folder_type));


    /* Must have be on connection cache */
    remove_from_cache(c,0);

    c->f = f;
    c->type->cache_folder_from_it(c,f);
}

void browser_from_connection(c,d)
     struct connection_cache *c;
     struct folder_browser *d;
{
    dprint(10, (debugfile,
		"browser_from_connection: con=%p (%s@%s), type=%p\n",
		c,
		c->C.username ? c->C.username : "<NULL>",
		c->C.host ? c->C.host : "<NULL>",
		c->type));
    dprint(10, (debugfile,
		"                      : browser=%p (%s), type=%p\n",
		d,d->sys_dir,d->type));

    /* Must have be on connection cache */
    remove_from_cache(c,0);

    c->d = d;
    c->type->cache_browser_from_it(c,d);
}

void  close_connection(con)
     struct connection_cache *con;
{
    dprint(10, (debugfile,
		"close_connection: con=%p (%s@%s), type=%p\n",
		con,
		con->C.username ? con->C.username : "<NULL>",
		con->C.host ? con->C.host : "<NULL>",
		con->type));
    
    con->type->cache_close_it(con);

    if (-1 != con->C.socket ) {
	dprint(10, (debugfile,
		    "close_connection: Closing socket\n"));
	
	clear_action(con->C.socket); /* Assurance */
	close(con->C.socket);        
	con->C.socket = -1;
    }	
    
    con->state = CON_error;
}

int close_cached_connections()
{
    struct connection_cache * ptr, *next;
    int X = 0;

    dprint(4, (debugfile,
		"Closing cached connections...\n"));
    for (ptr = CACHE_LIST; ptr; ptr = next) {
	next = ptr -> next;
	close_connection(ptr);
	X++;
	free_connection(&ptr);
    }
    dprint(4, (debugfile,
	       "Cached connections closed\n"));

    return X;
}

void  cache_connection(c)
     struct connection_cache *c;
{
    dprint(10, (debugfile,
		"cache_connection: con=%p (%s@%s), type=%p\n",
		c,
		c->C.username ? c->C.username : "<NULL>",
		c->C.host ? c->C.host : "<NULL>",
		c->type));
    
    c->f    = NULL;            /* Detach from folder */
    c->d    = NULL;            /* Detach from browser */
    c->next = CACHE_LIST;
    CACHE_LIST = c;    
}


#endif

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */


